from gevent.pool import Pool as BasePoll
import gevent.monkey
gevent.monkey.patch_all()


class Pool(BasePoll):
    '''routine pool'''

    def apply_async(self, func, args=None, kwds=None, callback=None):
        return BasePoll().apply_async(func, args=(), kwds={}, callback=callback)

    def close(self):
        pass



'''调度器模块'''
import six
import w3lib.url
from hashlib import sha1
from six.moves.queue import Queue

from elegance_spider.conf import settings
from elegance_spider.utils.queue import Queue as RedisQueue
from elegance_spider.conf.settings import SCHEDULER_PRESIST
from elegance_spider.set import NormalFilterContainer, RedisFilterContainer
from elegance_spider.utils.log import logger


class Scheduler(object):
    '''调度器类'''

    def __init__(self, collector):
        self._filter_container = set()
        if SCHEDULER_PRESIST:
            self.queue = RedisQueue()
            self._filter_container = RedisFilterContainer()
        else:
            self.queue = Queue()
            self._filter_container = NormalFilterContainer()
        self.collectot = collector
        self.repeate_request_num = 0

    def add_request(self, request):
        if request.filter is False:
            self.queue.put(request)
            logger.info("添加请求成功：<disable去重>[%s %s]" %
                        (request.method, request.url))
            self.total_request_number += 1
            return

        fp = self._gen_fp(request)
        if not self._filter_request(fp, request):
            self.queue.put(request)
            logger.info("添加请求成功：<去重>[%s %s]" % (request.method, request.url))
            if settings.ROLE in ['master', 'slave']:
                self._backup_request.save_request(fp, request)
            self._filter_container.add_fp(fp)
            self.total_request_number += 1
        else:
            logger.info("重复请求 [%s %s]" % (request.method, request.url))
            self.repeate_request_num += 1

    def get_request(self):
        try:
            request = self.queue.get(False)
        except:
            return None
        else:
            if request.filter is True and settings.ROLE in ['master', 'salve']:
                fp = self._gen_fp(request)
                if request.retry_time >= settings.MAX_RETRY_TIMES:
                    self._backup_request.delete_request(fp)
                    logger.warning("出现异常请求，且超过最大尝试次数：[%s]%s" % (
                        request.method, request.url))
                request.retry_time += 1
                self._backup_request.update_request(fp, request)
            return request

    def delete_request(self, request):
        if settings.ROLE in ['master', 'slave']:
            fp = self._gen_fp(request)
            self._backup_request.delete_request(fp)

    def add_lost_request(self):
        if settings.ROLE in ['master', 'slave']:
            for request in self._backup_request.get_requests():
                self.queue.put(request)

    def _filter_request(self, request):
        request.fp = self._gen_fp(request)
        if not self._filter_container.exists(request.fp):
            self._filter_container.add_fp(request.fp)
            return True
        else:
            logger.info("发现重复的请求：<%s>" % request.url)
            self.repeate_request_num += 1
            return False

    def _gen_fp(self, request):
        '''请求去重，计算指纹'''
        url = w3lib.url.canonicalize_url(request.url)
        method = request.method.upper()
        data = request.data if request.data is not None else {}
        data = sorted(data.items(), key=lambda x: x[0])

        s1 = sha1()
        s1.update(self._to_bytes(url))
        s1.update(self._to_bytes(method))
        s1.update(self._to_bytes(str(data)))

        fp = s1.hexdigest()
        return fp

    @staticmethod
    def _to_bytes(s):
        if six.PY2:
            if isinstance(s, str):
                return s
            else:
                return s.encode('utf-8')
        elif six.PY3:
            if isinstance(s, str):
                return s.encode('utf-8')
            else:
                return s



class DownloaderMiddleware(object):
    '''下载器中间件'''

    def process_request(self, request):
        '''预处理请求'''
        print('下载器中间件，预处理请求')
        return request

    def process_response(self, response):
        '''预处理响应'''
        print('下载器中间件，预处理响应')
        return response


class TestDownloaderMiddleware1(object):

    def process_request(self, request):
        '''处理请求头，添加默认的user-agent'''
        print("TestDownloaderMiddleware1: process_request")
        return request

    def process_response(self, item):
        '''处理数据对象'''
        print("TestDownloaderMiddleware1: process_response")
        return item


class TestDownloaderMiddleware2(object):

    def process_request(self, request):
        '''处理请求头，添加默认的user-agent'''
        print("TestDownloaderMiddleware2: process_request")
        return request

    def process_response(self, item):
        '''处理数据对象'''
        print("TestDownloaderMiddleware2: process_response")
        return item



"""Response对象封装"""
import json
import re

from lxml import etree


class Response(object):
    def __init__(self, url, status_code, headers, body, meta={}):
        self.url = url
        self.status_code = status_code
        self.headers = headers
        self.body = body
        self.meta = meta

    def xpath(self, rule):
        '''xpath解析'''
        html = etree.HTML(self.body)
        return html.xpath(rule)

    @property
    def json(self):
        '''json解析'''
        return json.loads(self.body)

    def re_findall(self, rule, data=None):
        '''findall正则'''
        if data is None:
            data = self.body
        return re.findall(rule, data)



'''队列'''
import time
import redis
import pickle
from six.moves import queue as BaseQueue

from elegance_spider.conf.settings import REDIS_QUEUE_HOST, REDIS_QUEUE_PORT, REDIS_QUEUE_DB, REDIS_QUEUE_NAME, REDIS_QUEUE_PASSWORD


class Queue(object):
    '''Redis as queue'''
    Empty = BaseQueue.Empty
    Full = BaseQueue.Full
    max_timeout = 0.3

    def __init__(self, maxsize=0, lazy_limit=True):
        self.name = REDIS_QUEUE_NAME
        self.redis = redis.StrictRedis(
            host=REDIS_QUEUE_HOST, port=REDIS_QUEUE_PORT, db=REDIS_QUEUE_DB, password=REDIS_QUEUE_PASSWORD)
        self.maxsize = maxsize
        self.lazy_limit = lazy_limit
        self.last_qsize = 0

    def qsize(self):
        self.last_qsize = self.redis.llen(self.name)
        return self.last_qsize

    def empty(self):
        if self.qsize() == 0:
            return True
        else:
            return False

    def full(self):
        if self.maxsize and self.qsize() >= self.maxsize:
            return True
        else:
            return False

    def put_nowait(self, obj):
        if self.lazy_limit and self.last_qsize < self.maxsize:
            pass
        elif self.full():
            raise self.Full
        self.last_qsize = self.redis.rpush(self.name, pickle.dumps(obj))
        return True

    def put(self, obj, block=True, timeout=None):
        if not block:
            return self.put_nowait(obj)

        start_time = time.time()
        while True:
            try:
                return self.put_nowait(obj)
            except self.Full:
                if timeout:
                    lasted = time.time() - start_time
                    if timeout > lasted:
                        time.sleep(min(self.max_timeout, timeout - lasted))
                    else:
                        raise
                else:
                    time.sleep(self.max_timeout)

    def get_nowait(self):
        ret = self.redis.lpop(self.name)
        if ret is None:
            raise self.Empty
        return pickle.loads(ret)

    def get(self, block=True, timeout=None):
        if not block:
            return self.get_nowait()

        start_time = time.time()
        while True:
            try:
                return self.get_nowait()
            except self.Empty:
                if timeout:
                    lasted = time.time() - start_time
                    if timeout > lasted:
                        time.sleep(min(self.max_timeout, timeout - lasted))
                    else:
                        raise
                else:
                    time.sleep(self.max_timeout)



class SpiderMiddleware(object):
    '''爬虫中间件'''

    def process_request(self, request):
        '''请求处理'''
        print('爬虫中间件')
        return request

    def process_response(self, response):
        '''预处理数据'''
        print('爬虫中间件预处理')
        return response


class TestSpiderMiddleware1(object):
    ''''''

    def process_request(self, request):
        print('TestSpiderMiddleware1: process_request')
        return request

    def process_item(self, item):
        print('TestSpiderMiddleware1: process_item')
        return item


class TestSpiderMiddleware2(object):
    ''''''

    def process_request(self, request):
        print('TestSpiderMiddleware2: process_request')
        return request

    def process_item(self, item):
        print('TestSpiderMiddleware2: process_item')
        return item



"""Item对象"""


class Item(object):
    """
    Item对象
    """

    def __init__(self, data):
        self._data = data

    @property
    def data(self):
        '''对外提供data访问'''
        return self._data



'''爬虫引擎'''
import time
import datetime
import importlib

from elegance_spider.http.request import Request
from elegance_spider.utils.log import logger
from elegance_spider.utils.stats_collector import StatsColector

from .scheduler import Scheduler
from .downloader import Downloader

from elegance_spider.conf.settings import SPIDERS, PIPLINES, SPIDER_MIDDLEWARES, DOWNLOADER_MIDDLEWARES, MAX_ASYNC_NUMBER, ASYNC_TYPE


if ASYNC_TYPE == 'thread':
    from multiprocessing.dummy import Pool
elif ASYNC_TYPE == 'coroutine':
    from elegance_spider.utils.coroutine import Pool
else:
    raise Exception('不支持的异步类型，%s, 只能支持thread和coroutine' % ASYNC_TYPE)


class Engine(object):

    def __init__(self):
        self.spiders = self._auto_import_instances(SPIDERS, isspider=True)
        self.downloader = Downloader()
        self.piplines = self._auto_import_instances(PIPLINES)
        self.collector = StatsColector(self.spiders)
        self.scheduler = Scheduler(self.collector)
        self.spider_middlewares = self._auto_import_instances(
            SPIDER_MIDDLEWARES)
        self.downloader_middlewares = self._auto_import_instances(
            DOWNLOADER_MIDDLEWARES)
        self.pool = Pool(4)
        self.running = False

    def _auto_import_instances(self, path=[], isspider=False):
        if isspider is True:
            instances = {}
        else:
            instances = []
        for p in path:
            module_name = p.rsplit(".", 1)[0]
            cls_name = p.rsplit(".", 1)[-1]
            model = importlib.import_module(module_name)
            cls = getattr(model, cls_name)
            if isspider is True:
                instances[cls_name] = cls()
            else:
                instances.append(cls())
        return instances

    def start(self):
        start_time = datetime.datetime.now()
        logger.info('爬虫开始：{}' . format(start_time))
        logger.info('爬虫运行模式：{}' . format(ASYNC_TYPE))
        logger.info('最大并发数：{}' . format(MAX_ASYNC_NUMBER))
        logger.info('启动的爬虫：{}' . format(list(self.spiders.keys())))
        logger.info('启动的下载中间件：\n{}' . format(DOWNLOADER_MIDDLEWARES))
        logger.info('启动的爬虫中间件：\n{}' . format(SPIDER_MIDDLEWARES))
        logger.info('启动的爬虫管道：\n{}' . format(PIPLINES))
        self._start_engine()
        end_time = datetime.datetime.now()
        logger.info("爬虫结束: {}" . format(end_time))
        logger.info("耗时：%s 秒" % (end_time - start_time).total_seconds())
        logger.info("总请求量: {}" . format(self.collector.request_nums))
        logger.info("重复的请求: {}个" . format(self.collector.repeat_request_nums))
        logger.info("成功的请求: {}个" . format(self.collector.response_nums))
        self.collector.clear()

    def _start_request_callback(self, temp):
        self.collector.incr(self.collector.start_request_nums_key)

    def _start_request(self):
        def _func(spider_name, spider):
            for start_request in spider.start_requests():
                for spider_middleware in self.spider_middlewares:
                    start_request = spider_middleware.process_request(
                        start_request)
                start_request.spider_name = spider_name
                self.scheduler.add_request(start_request)
                self.collector.incr(self.collector.request_nums_key)
        for spider_name, spider in self.spiders.items():
            self.pool.apply_async(_func, args=(
                spider_name, spider), callback=self._start_request_callback)

    def _execute_request_response_item(self):
        request = self.scheduler.get_request()
        if request is None:
            return

        spider = self.spiders[request.spider_name]

        for downloader_middleware in self.downloader_middlewares:
            request = downloader_middleware.process_request(request)

        response = self.downloader.get_response(request)

        response.meta = request.meta

        for downloader_middleware in self.downloader_middlewares:
            response = downloader_middleware.process_response(response)

        for spider_middleware in self.spider_middlewares:
            response = spider_middleware.process_response(response)

        parse = getattr(spider, request.parse)

        for result in parse(response):
            if isinstance(result, Request):
                result.spider_name = request.spider_name
                for spider_middleware in self.spider_middlewares:
                    result = spider_middleware.process_request(result)
                self.scheduler.add_request(result)
                self.collector.incr(self.collector.request_nums_key)
            else:
                for pipline in self.piplines:
                    result = pipline.process_item(result, spider)
        self.collector.incr(self.collector.response_nums_key)

    def _callback(self, _):
        if self.running:
            self.pool.apply_async(
                self._execute_request_response_item, callback=self._callback)

    def _start_engine(self):
        self.running = True
        self.pool.apply_async(self._start_request, callback=self._callback)
        for i in range(MAX_ASYNC_NUMBER):
            self.pool.apply_async(
                self._execute_request_response_item, callback=self._callback)

        while True:
            time.sleep(0.001)
            if self.collector.response_nums + self.collector.repeat_request_nums >= self.collector.request_nums:
                self.running = False
                break

        self.pool.close()
        self.pool.join()



'''下载器'''

import requests
from elegance_spider.http.response import Response


class Downloader(object):
    def get_response(self, request):
        if request.method.upper() == 'GET':
            resp = requests.get(request.url, headers=request.headers, params=request.params)
        elif request.method.upper() == 'POST':
            resp = requests.post(request.url, headers=request.headers, params=request.param, data=request.data)
        else:
            raise Exception('不支持的请求方法')

        return Response(resp.url, resp.status_code, resp.headers, resp.content)


import redis
import pickle

from elegance_spider.http.request import Request
from elegance_spider.conf import settings


class RedisBackupRequest(object):
    '''利用hash类型，存储每个请求对象，key是指纹，value是请求对象'''
    REDIS_BACKUP_NAME = settings.REDIS_BACKUP_NAME
    REDIS_BACKUP_HOST = settings.REDIS_BACKUP_HOST
    REDIS_BACKUP_PORT = settings.REDIS_BACKUP_PORT
    REDIS_BACKUP_DB = settings.REDIS_BACKUP_DB
    REDIS_BACKUP_PASSWORD = settings.REDIS_BACKUP_PASSWORD

    def __init__(self):
        self._redis = redis.StrictRedis(
            host=self.REDIS_BACKUP_HOST, port=self.REDIS_BACKUP_PORT, db=self.REDIS_BACKUP_DB, password=self.REDIS_BACKUP_PASSWORD)
        self._name = self.REDIS_BACKUP_NAME

    def save_request(self, fp, request):
        bytes_data = pickle.dumps(request)
        self._redis.hset(self._name, fp, bytes_data)

    def delete_request(self):
        self._redis.hdel(self._name, fp)

    def update_request(self, fp, request):
        self.save_request(fp, request)

    def get_requests(self):
        for _, bytes_request in self._redis.hscan_iter(self._name):
            request = pickle.loads(bytes_request)
            yield request



'''爬虫组件'''
from elegance_spider.item import Item
from elegance_spider.http.request import Request


class Spider(object):
    '''
    '''
    start_urls = []

    def start_requests(self):
        '''构建初始请求对象并返回'''
        for url in self.start_urls:
            yield Request(url)

    def parse(self, response):
        '''解析响应对象'''
        yield Item(response.body)



import logging

DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FMT = '%(asctime)s %(filename)s[line:%(lineno)d] \
                  %(levelname)s: %(message)s'
DEFAULT_LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'
DEFAULT_LOG_FILENAME = 'log.log'

SPIDERS = []
PIPLINES =[]
SPIDER_MIDDLEWARES = []
DOWNLOADER_MIDDLEWARES = []


"""封装Request对象"""


class Request(object):
    """
    Request对象
    """

    def __init__(self, url, method='GET', headers=None, params=None, data=None, parse='parse', filter=True):
        self.url = url
        self.method = method
        self.headers = headers
        self.data = data
        self.params = params
        self.parse = parse
        self.meta = None
        self.filter = filter
        self.retry_time = 0



from .default_settings import *

SCHEDULER_PRESIST = False
ASYNC_TYPE = 'coroutine'
MAX_ASYNC_NUMBER = 2

# redis queue
REDIS_QUEUE_NAME = 'spider_queue'
REDIS_QUEUE_HOST = 'localhost'
REDIS_QUEUE_PORT = 6379
REDIS_QUEUE_DB = 2
REDIS_QUEUE_PASSWORD = None

# set
REDIS_SET_NAME = 'spider_set'
REDIS_SET_HOST = 'localhost'
REDIS_SET_PORT = 6379
REDIS_SET_DB = 2
REDIS_SET_PASSWORD = None

# backup
REDIS_BACKUP_NAME = 'spider_backup'
REDIS_BACKUP_HOST = 'localhost'
REDIS_BACKUP_PORT = 6379
REDIS_BACKUP_DB = 2
REDIS_BACKUP_PASSWORD = None

# role
ROLE = 'master'


try:
    from settings import *
except:
    pass




import sys
import logging

from elegance_spider.conf import settings


class Logger(object):
    def __init__(self):
        self._logger = logging.getLogger()
        self.formatter = logging.Formatter(
            fmt=settings.DEFAULT_LOG_FMT, datefmt=settings.DEFAULT_LOG_DATEFMT)
        self._logger.addHandler(self._get_file_handler(
            settings.DEFAULT_LOG_FILENAME))
        self._logger.addHandler(self._get_console_handler())
        self._logger.setLevel(settings.DEFAULT_LOG_LEVEL)

    def _get_file_handler(self, filename):
        filehandler = logging.FileHandler(filename=filename, encoding='utf-8')
        filehandler.setFormatter(self.formatter)
        return filehandler

    def _get_console_handler(self):
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.formatter = self.formatter
        return console_handler

    @property
    def logger(self):
        return self._logger


logger = Logger().logger



from os.path import dirname, join

from setuptools import (
    find_packages,
    setup,
)


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


with open(join(dirname(__file__), 'VERSION.txt'), 'rb') as f:
    version = f.read().decode('ascii').strip()

setup(
    name='elegance_spider',
    version=version,
    description='A spider framework',
    packages=find_packages(exclude=[]),
    author='kirintang',
    author_email='215936564@qq.com',
    license='MIT',
    package_data={'': ['*.*']},
    url='https://kirintang.github.io',
    install_requires=parse_requirements("requirements.txt"),
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



import redis
from elegance_spider.conf.settings import REDIS_QUEUE_HOST, REDIS_QUEUE_PORT, REDIS_QUEUE_DB, REDIS_QUEUE_PASSWORD


class StatsColector(object):

    def __init__(self, spider_names=[]):
        self.redis = redis.StrictRedis(
            host=REDIS_QUEUE_HOST, port=REDIS_QUEUE_PORT, db=REDIS_QUEUE_DB, password=REDIS_QUEUE_PASSWORD)
        self.request_nums_key = "_" . join(spider_names) + "_request_nums"
        self.response_nums_key = "_" . join(spider_names) + "_response_nums"
        self.repeat_request_nums_key = "_" . join(
            spider_names) + "_repeat_request_nums"
        self.start_request_nums_key = "_" . join(
            spider_names) + "_start_request_nums"

    def incr(self, key):
        self.redis.incr(key)

    def get(self, key):
        ret = self.redis.get(key)
        if not ret:
            ret = 0
        else:
            ret = int(ret)
        return ret

    def clear(self):
        self.redis.delete(self.request_nums_key, self.response_nums_key,
                          self.repeat_request_nums_key, self.start_request_nums_key)

    @property
    def request_nums(self):
        return self.get(self.request_nums_key)

    @property
    def response_nums(self):
        return self.get(self.response_nums_key)

    @property
    def repeat_request_nums(self):
        return self.get(self.repeat_request_nums_key)

    @property
    def start_request_nums(self):
        return self.get(self.start_request_nums_key)



import redis
from elegance_spider.conf import settings


class BaseFilterContainer(object):

    def add_fp(self, fp):
        pass

    def exists(self, fp):
        pass


class NormalFilterContainer(BaseFilterContainer):
    def __init__(self):
        self._filter_container = set()

    def add_fp(self, fp):
        self._filter_container.add(fp)

    def exists(self, fp):
        if fp in self._filter_container:
            return True
        else:
            return False


class RedisFilterContainer(BaseFilterContainer):
    REDIS_SET_NAME = settings.REDIS_SET_NAME
    REDIS_SET_HOST = settings.REDIS_SET_HOST
    REDIS_SET_PORT = settings.REDIS_SET_PORT
    REDIS_SET_DB = settings.REDIS_SET_DB

    def __init__(self):
        self._redis = redis.StrictRedis(
            host=self.REDIS_SET_HOST, port=self.REDIS_SET_PORT, db=self.REDIS_SET_DB)
        self._name = self.REDIS_SET_NAME

    def add_fp(self, fp):
        self._redis.sadd(self._name, fp)

    def exists(self, fp):
        return self._redis.sismember(self._name, fp)
